using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public enum SBPrefStatus {
	NOT_REQUESTED,
	READY,
	BUSY_LOADING,
	ERROR
}

public static class SBPreferences {

	private static string ServerName= "http://www.scatterbrain.cc/DBPortal/SBPreferences.php";
	private static string HashPassword= "SECRET";

	private static float lastCallTime= -100f;  // last call to server, try to restrict calls
	private static float lastCallTimeOut= 20f;
	private static string message= "";         // last message

	private static string keyName;
	private static Dictionary<string, object> dictionary= new Dictionary<string, object>();
	private static SBPrefStatus preferenceStatus= SBPrefStatus.NOT_REQUESTED;  // false=not ready, true=ready, null=error/not requested
	private static bool dirty= false;
	
	public static MonoBehaviour monobehaviour= null;
	public static MonoBehaviour m {
		get {
			if (SBPreferences.monobehaviour == null) {
				GameObject go= new GameObject("zzzSBPreferenceConnection");
				SBPreferences.monobehaviour= (MonoBehaviour)go.AddComponent(typeof(MonoBehaviour));
				if (SBPreferences.monobehaviour == null) {
					Debug.Log("ERROR - create score getter");
				} else {
					GameObject.DontDestroyOnLoad(go);
				}
			}
			return SBPreferences.monobehaviour;
		}
    }
	

	// BOOL
	public static bool GetBool(string key) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "B")) {
			return (bool)dictionary[key + "B"];
		}
		return false;
	}
	public static bool GetBool(string key, bool defaultVal) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "B")) {
			return (bool)dictionary[key + "B"];
		}
		return defaultVal;
	}
	public static void SetBool(string key, bool value) {
		if (key.Length > 0) {
			if (dictionary.ContainsKey(key + "B")) {
				dictionary.Remove(key + "B");
			}
			SBPreferences.dirty= true;
			dictionary.Add(key + "B", value);
		}
	}

	// INT
	public static int GetInt(string key) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "I")) {
			return (int)dictionary[key + "I"];
		}
		return 0;
	}
	public static int GetInt(string key, int defaultVal) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "I")) {
			return (int)dictionary[key + "I"];
		}
		return defaultVal;
	}
	public static void SetInt(string key, int value) {
		if (key.Length > 0) {
			if (dictionary.ContainsKey(key + "I")) {
				dictionary.Remove(key + "I");
			}
			SBPreferences.dirty= true;
			dictionary.Add(key + "I", value);
		}
	}
	
	// FLOAT
	public static float GetFloat(string key) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "F")) {
			return (float)dictionary[key + "F"];
		}
		return 0f;
	}
	public static float GetFloat(string key, float defaultVal) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "F")) {
			return (float)dictionary[key + "F"];
		}
		return defaultVal;
	}
	public static void SetFloat(string key, float value) {
		if (key.Length > 0) {
			if (dictionary.ContainsKey(key + "F")) {
				dictionary.Remove(key + "F");
			}
			SBPreferences.dirty= true;
			dictionary.Add(key + "F", value);
		}			
	}
	
	// STRING
	public static string GetString(string key) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "S")) {
			return (string)dictionary[key + "S"];
		}
		return "";
	}
	public static string GetString(string key, string defaultVal) {
		if (key.Length > 0 && dictionary.ContainsKey(key + "S")) {
			return (string)dictionary[key + "S"];
		}
		return defaultVal;
	}
	public static void SetString(string key, string value) {
		if (key.Length > 0) {
			if (dictionary.ContainsKey(key + "S")) {
				dictionary.Remove(key + "S");
			}
			SBPreferences.dirty= true;
			dictionary.Add(key + "S", value);
		}
	}

	
	// HAS KEY
	public static bool HasKey(string key) {
		return (dictionary.ContainsKey(key + "B") || dictionary.ContainsKey(key + "I") || dictionary.ContainsKey(key + "F") || dictionary.ContainsKey(key + "S"));
	}

	// DELETE KEY
	public static void DeleteKey(string key) {
		if (dictionary.ContainsKey(key + "B")) {
			dictionary.Remove(key + "B");
			SBPreferences.dirty= true;
		}
		if (dictionary.ContainsKey(key + "I")) {
			dictionary.Remove(key + "I");
			SBPreferences.dirty= true;
		}
		if (dictionary.ContainsKey(key + "F")) {
			dictionary.Remove(key + "F");
			SBPreferences.dirty= true;
		}
		if (dictionary.ContainsKey(key + "S")) {
			dictionary.Remove(key + "S");
			SBPreferences.dirty= true;
		}
	}
	
	// DELETE ALL
	public static void DeleteAll(string key) {
		SBPreferences.dirty= true;
		dictionary.Clear();
	}
	
	
	// RESET STATUS
	public static void ResetStatus() {
		SBPreferences.preferenceStatus= SBPrefStatus.NOT_REQUESTED;
		SBPreferences.message= "";
	}
	
	
	////////////////////////////////
	// ACCESSORS
	////////////////////////////////

	public static string Message {
		get { return SBPreferences.message; }
	}

	public static SBPrefStatus Status {
		get { return SBPreferences.preferenceStatus; }
	}


	////////////////////////////////
	// DATABASE CALLS
	////////////////////////////////



	// PREFERENCES READY
	// get the data, but you can call repeatedly until they are ready
	// ret: false if not ready, true if success getting data, null if failure
	public static SBPrefStatus ReadPreferences(string key) {
		if (key == "") {
			SBPreferences.message= "Invalid Empty Preference Key";
			return SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
		}
		if (SBPreferences.preferenceStatus == SBPrefStatus.BUSY_LOADING) {  // if we are already waiting on a request
			return SBPrefStatus.BUSY_LOADING;
		} else if (!SBPreferences.dirty && key == SBPreferences.keyName && SBPreferences.preferenceStatus == SBPrefStatus.READY) {  // we heard back
			SBPreferences.message= "Nothing Read - Using Local Preferences";
			return SBPrefStatus.READY;
		} else if (SBPreferences.dirty || (Time.time - SBPreferences.lastCallTime) > SBPreferences.lastCallTimeOut) {
			if (SBPreferences.dirty) {
				Debug.Log("WARNING - writing over local values");
			}
			SBPreferences.m.StartCoroutine(SBPreferences.ReadPreferencesCoroutine(keyName));  // kick off the routine to contact server
			return SBPrefStatus.BUSY_LOADING;
		}
		return SBPrefStatus.BUSY_LOADING;
	}


	// READY PREFERENCES
	// Call DB to get dictionary.  Will always make call to db
	public static IEnumerator ReadPreferencesCoroutine(string key) {
		if (key == "") {
			SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
			SBPreferences.message= "Invalid Empty Preference Key";
			yield break;
		}
		if (SBPreferences.preferenceStatus == SBPrefStatus.BUSY_LOADING) {
			yield break;			
		}
		SBPreferences.lastCallTime= Time.time;
		SBPreferences.preferenceStatus= SBPrefStatus.BUSY_LOADING;
		SBPreferences.message= "";
		SBPreferences.keyName= key;

		WWW dataSite = new WWW(SBPreferences.ServerName + "?Req=GET&Key=" + WWW.EscapeURL(key));
		while (!dataSite.isDone) {
			if ((Time.time - SBPreferences.lastCallTime) > SBPreferences.lastCallTimeOut) {
				SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
				SBPreferences.message= "Timeout";
				yield break;
			}
			yield return 10;
		}

		try {
			if (dataSite.error == null) {
				if (dataSite.text.Length == 0) {
					SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
					SBPreferences.message= "NO DATA";
				} else if ((SBPreferences.dictionary= SBPreferences.DecodeDictionary(dataSite.text)) != null) {
					SBPreferences.preferenceStatus= SBPrefStatus.READY;
					SBPreferences.message= "SUCCESS";
					SBPreferences.dirty= false;
				}
			} else {
				SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
				SBPreferences.message= dataSite.error;
			}
		} catch (System.Exception e) {
			SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
			SBPreferences.message= "Conversion Error: " + e.ToString();
		}
		
		yield break;
	}


	// WRITE PREFERENCES
	// Write the current preferences to the database
	public static void WritePreferences(string key) {
		if (key == "") {
			SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
			SBPreferences.message= "Invalid Empty Preference Key";
			return;
		}
		
		if (SBPreferences.dirty && SBPreferences.preferenceStatus != SBPrefStatus.BUSY_LOADING || (Time.time - SBPreferences.lastCallTime) > SBPreferences.lastCallTimeOut) {
			SBPreferences.m.StartCoroutine(SBPreferences.WritePreferencesCoroutine(key));
		}
	}
	
	
	// WRITE PREFERENCES
	// Sends values into server and waits for return
	public static IEnumerator WritePreferencesCoroutine(string key) {
		if (key == "") {
			SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
			SBPreferences.message= "Invalid Empty Preference Key";
			yield break;
		}
		if (SBPreferences.preferenceStatus == SBPrefStatus.BUSY_LOADING) {
			yield break;			
		}

		SBPreferences.lastCallTime= Time.time;
		SBPreferences.preferenceStatus= SBPrefStatus.BUSY_LOADING;
		SBPreferences.message= "";
		SBPreferences.keyName= key;

		string encoded= SBPreferences.EncodeDictionary(dictionary);
		string chksum= SBPreferences.ChkSum(key + encoded.Substring(10, 30));
		WWW dataSite = new WWW(
			SBPreferences.ServerName + "?Req=PUT&Key=" + WWW.EscapeURL(key) + "&CHK=" + chksum,
			System.Text.Encoding.ASCII.GetBytes("data=" + encoded)
		);
		
		float timeout= Time.time;
		while (!dataSite.isDone) {
			if ((Time.time - timeout) > 10f) {
				SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
				SBPreferences.message= "Timeout";
				yield break;
			}
			yield return 10;
		}
		if (dataSite.error == null) {
			if (dataSite.text.Length > 0 && dataSite.text[dataSite.text.Length-1] == '0') {
				SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
				SBPreferences.message= dataSite.text;
			} else {
				SBPreferences.preferenceStatus= SBPrefStatus.READY;
				SBPreferences.message= "SUCCESS";
				SBPreferences.dirty= false;
			}
		} else {
			SBPreferences.preferenceStatus= SBPrefStatus.ERROR;
			SBPreferences.message= dataSite.error;
		}
		yield break;
	}


	// CHKSUM
	// Calculate a Checksum
	public static string ChkSum(string inStr) {
		inStr+= SBPreferences.HashPassword;
		
		System.Text.Encoding enc= new System.Text.ASCIIEncoding();
		byte[] buffer= enc.GetBytes(inStr);
		System.Security.Cryptography.SHA1CryptoServiceProvider cryptoTransformSHA1=
			new System.Security.Cryptography.SHA1CryptoServiceProvider();
		return System.BitConverter.ToString(cryptoTransformSHA1.ComputeHash(buffer)).Replace("-", "");
	}



	////////////////////////////////
	// ENCODING / DECODING
	////////////////////////////////
	public static string EncodeDictionary(Dictionary<string,object> dataDict) {
		System.IO.MemoryStream ms= new System.IO.MemoryStream();
		System.Runtime.Serialization.Formatters.Binary.BinaryFormatter bFormatter =
			new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();
		
		bFormatter.Serialize(ms, dataDict);

		string dataString= System.Convert.ToBase64String(ms.ToArray()).Replace("+", "_");
		return dataString;
	}
	public static Dictionary<string,object> DecodeDictionary(string dataString) {
	
		dataString= dataString.Replace("_","+");
		System.IO.MemoryStream ms= new System.IO.MemoryStream(System.Convert.FromBase64String(dataString));

		System.Runtime.Serialization.Formatters.Binary.BinaryFormatter bFormatter = 
			new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();
		return (Dictionary<string,object>)bFormatter.Deserialize(ms);
	}
}
